"use strict";

var _interopRequireWildcard = require("babel-runtime/helpers/interop-require-wildcard")["default"];

exports.__esModule = true;
exports.WithStatement = WithStatement;
exports.IfStatement = IfStatement;
exports.ForStatement = ForStatement;
exports.WhileStatement = WhileStatement;
exports.DoWhileStatement = DoWhileStatement;
exports.LabeledStatement = LabeledStatement;
exports.TryStatement = TryStatement;
exports.CatchClause = CatchClause;
exports.SwitchStatement = SwitchStatement;
exports._switchCase = _switchCase;
exports.SwitchCase = SwitchCase;
exports.SafeSwitchCase = SafeSwitchCase;
exports.DebuggerStatement = DebuggerStatement;
exports.VariableDeclaration = VariableDeclaration;
exports.VariableDeclarator = VariableDeclarator;

var _babelTypes = require("babel-types");

var t = _interopRequireWildcard(_babelTypes);

function WithStatement(node) {
  this.keyword("with");
  this.print(node, "object");
  this.printBlock(node);
}

function IfStatement(node) {
  this.keyword("if");
  this.print(node, "test");

  this.printBlock(node, "consequent");
  if (node.alternate) {
    this.keyword("else");
    this.printBlock(node, "alternate");
  }
}

function ForStatement(node) {
  this.keyword("for");

  if (node.init) this.print(node, "init");

  if (node.test) {
    this.keyword("while");
    this.print(node, "test");
  }

  if (node.update) {
    this.keyword("update");
    this.print(node, "update");
  }

  this.printBlock(node);
}

// TODO: forUptoStatement
// TODO: forDowntoStatement

function WhileStatement(node) {
  this.keyword("while");
  this.print(node, "test");
  this.printBlock(node);
}

var buildForXStatement = function buildForXStatement(op) {
  return function (node) {
    this.keyword("for");
    this.print(node, "left");
    this.keyword(op);
    this.print(node, "right");
    this.printBlock(node);
  };
};

var ForInStatement = buildForXStatement("in");
exports.ForInStatement = ForInStatement;
var ForOfStatement = buildForXStatement("of");

exports.ForOfStatement = ForOfStatement;

function DoWhileStatement(node) {
  this.push("do");
  this.printBlock(node);
  this.keyword("while");
  this.print(node, "test");
  this.newline();
}

// TODO: if syntax is ambiguous, use a !
function buildLabelStatement(prefix) {
  var key = arguments.length <= 1 || arguments[1] === undefined ? "label" : arguments[1];

  return function (node) {
    this.push(prefix);

    if (node[key]) {
      this.print(node, key);
    }
    this.newline();
  };
}

var ContinueStatement = buildLabelStatement("continue");
exports.ContinueStatement = ContinueStatement;
var ReturnStatement = buildLabelStatement("return", "argument");
exports.ReturnStatement = ReturnStatement;
var BreakStatement = buildLabelStatement("break");
exports.BreakStatement = BreakStatement;
var ThrowStatement = buildLabelStatement("throw", "argument");

exports.ThrowStatement = ThrowStatement;

function LabeledStatement(node) {
  this.print(node, "label");
  this.push(":");
  this.print(node, "body");
}

function TryStatement(node) {
  this.keyword("try");
  this.printBlock(node, "block");

  if (node.handler) this.print(node, "handler");

  if (node.finalizer) {
    this.keyword("finally");
    this.printBlock(node, "finalizer");
  }
  if (!this.format.preserve) this.newline(true);
}

function CatchClause(node) {
  this.keyword("catch");
  this.print(node, "param");
  this.printBlock(node);
}

// TODO: SafeSwitchStatement

function SwitchStatement(node) {
  this.keyword("switch");
  this.push("!");
  this.print(node, "discriminant");
  this.newline();

  this.printStatements(node, "cases", {
    indent: true
  });

  if (!this.format.preserve) this.newline(true);
}

function _switchCase(node) {
  if (node.test) {
    this.keyword("case");
    this.print(node, "test");
  } else {
    this.keyword("default");
  }
}

function SwitchCase(node) {
  this._switchCase(node);
  if (node.consequent.length) {
    this.newline();
    this.printStatements(node, "consequent", { indent: true });
  }
  this.newline();
}

function SafeSwitchCase(node) {
  if (node.fallthrough) {
    this.keyword("and");
  }
  this._switchCase(node);
  // NOTE: this can't be a block, for scope reasons.
  // if collapse, if the array only contains one statement, it will be prefixed
  // with "then" and will have no leading newline.
  this.printStatements(node, "consequent", { indent: true, collapse: true });
}

function DebuggerStatement() {
  this.keyword("debugger");
  this.newline();
}

function VariableDeclaration(node, parent) {
  this.keyword(node.kind);

  var hasInits = false;
  // don't add whitespace to loop heads
  var isLoopHead = t.isFor(parent);
  if (!isLoopHead) {
    var _arr = node.declarations;

    for (var _i = 0; _i < _arr.length; _i++) {
      var declar = _arr[_i];
      if (declar.init) {
        // has an init so let's split it up over multiple lines
        hasInits = true;
      }
    }
  }
  if (this.format.preserve && node.tokenElements && node.tokenElements.length) {
    throw new Error("Not Implemented");
  } else {
    var useNewlines = !this.format.compact && hasInits && node.declarations.length > 1;
    var sep = useNewlines ? { type: "newline" } : ",";
    if (useNewlines) this.newline();
    this.printMultiple(node, "declarations", { separator: sep, indent: useNewlines });
    if (!isLoopHead) this.newline();
  }
}

function VariableDeclarator(node) {
  this.print(node, "id", { assertChildrenPrinted: ["typeAnnotation"] });
  if (node.init) {
    this.push("=");
    this.print(node, "init");
  }
}