
/**
 * Tacoscript's code generator, turns an ast into tacoscript code
 */

"use strict";

var _classCallCheck = require("babel-runtime/helpers/class-call-check")["default"];

var _interopRequireDefault = require("babel-runtime/helpers/interop-require-default")["default"];

exports.__esModule = true;

var _sourceMap = require("source-map");

var _sourceMap2 = _interopRequireDefault(_sourceMap);

var _position = require("./position");

var _position2 = _interopRequireDefault(_position);

var _horchata = require("horchata");

var _lodashLangIsString = require("lodash/lang/isString");

var _lodashLangIsString2 = _interopRequireDefault(_lodashLangIsString);

var _lodashLangIsEqual = require("lodash/lang/isEqual");

var _lodashLangIsEqual2 = _interopRequireDefault(_lodashLangIsEqual);

require("./special-tokens");

// TODO: move to tacoscript-options package
function normalizeOptions(input) {
  if (input == null) throw new Error("Unexpected null or undefined options");
  var inputDialect = input.dialect || {};
  // TODO: console.warn about unknown opts
  var opts = {};
  opts.dialect = {};
  opts.dialect["equality-symbols"] = !!inputDialect["equality-symbols"];
  opts.format = input.format || {};
  return opts;
}

var TacoBuffer = (function () {
  function TacoBuffer(opts, code) {
    _classCallCheck(this, TacoBuffer);

    this._initSourceMap(opts, code);
    opts = normalizeOptions(opts);
    this.opts = opts;
    this.format = opts.format;
    this.dialect = opts.dialect;
    this.tokens = [new _horchata.Token(_horchata.tokTypes.tab, 0)];
    this._indent = 0;
    this._lastIndent = 0;

    // serialization state
    this.position = new _position2["default"]();
    this.code = code;
    this.output = '';

    this._warnings = [];
  }

  /**
   * Checks
   */

  TacoBuffer.prototype.isLast = function isLast(state) {
    state = TacoBuffer._massageTokenState(state);
    var last = this._last();
    return last.type === state.type && _lodashLangIsEqual2["default"](last.value, state.value);
  };

  TacoBuffer.prototype.isLastType = function isLastType(type) {
    if (this.tokens.length <= 0) return false;
    var last = this._last();
    return last.type === type;
  };

  TacoBuffer.prototype._lastOffset = function _lastOffset() {
    var offset = arguments.length <= 0 || arguments[0] === undefined ? 1 : arguments[0];

    var i = offset;
    for (var len = this.tokens.length; i <= len && this.tokens[len - i].type === _horchata.tokTypes.mappingMark; i++) {}
    return i;
  };

  TacoBuffer.prototype._last = function _last() {
    return this.tokens[this.tokens.length - this._lastOffset()];
  };

  TacoBuffer.prototype._pop = function _pop() {
    var i, len, token;
    for (i = 1, len = this.tokens.length; i <= len && (token = this.tokens[len - i]).type === _horchata.tokTypes.mappingMark; i++) {}
    this.tokens.splice(len - i, 1);
    return token;
  };

  /**
   * Source Map
   */

  TacoBuffer.prototype._initSourceMap = function _initSourceMap(opts, code) {
    if (opts.sourceMaps) {
      if (!opts.sourceFileName) {
        throw new Error("sourceFileName must be set when generating source maps");
      }
      this.map = new _sourceMap2["default"].SourceMapGenerator({
        file: opts.sourceMapTarget,
        sourceRoot: opts.sourceRoot
      });
      this.map.setSourceContent(opts.sourceFileName, code);
    } else {
      this.map = null;
    }
  };

  TacoBuffer.prototype.mark = function mark(original) {
    if (!this.map) return;
    // use current location to mark the source map
    var position = this.position;
    this.map.addMapping({
      source: this.opts.sourceFileName,
      generated: {
        line: position.line,
        column: position.column
      },
      original: original
    });
  };

  /**
   * Tokenization
   */

  TacoBuffer.prototype.newline = function newline(force) {
    if (force || !this.isLastType(_horchata.tokTypes.newline)) this._push({ type: _horchata.tokTypes.newline });
  };

  TacoBuffer.prototype.indent = function indent() {
    // TODO: make sure that indent only occurs right before a newline
    this._indent++;
    if (this.isLastType(_horchata.tokTypes.newline)) {
      var newline = this._pop();
      this._insertIndentTokens();
      this.tokens.push(newline);
    }
  };

  TacoBuffer.prototype.dedent = function dedent() {
    this._indent--;
    if (this.isLastType(_horchata.tokTypes.newline)) {
      var newline = this._pop();
      this._insertIndentTokens();
      this.tokens.push(newline);
    }
  };

  TacoBuffer.prototype.keyword = function keyword(name) {
    this._push({ type: _horchata.keywordTypes[name] });
  };

  TacoBuffer.prototype.push = function push() {
    for (var _len = arguments.length, _tokenStates = Array(_len), _key = 0; _key < _len; _key++) {
      _tokenStates[_key] = arguments[_key];
    }

    var tokenStates = _tokenStates; // babel#2539
    var _arr = tokenStates;
    for (var _i = 0; _i < _arr.length; _i++) {
      var token = _arr[_i];
      this._push(token);
    }
  };

  TacoBuffer.prototype.flush = function flush() {
    var lastOffset = this._lastOffset();
    if (this.tokens[this.tokens.length - lastOffset].type !== _horchata.tokTypes.newline) {
      this.onWarning("Last token is not a newline");
      var last = this._last();
      if (last != null && last.type === _horchata.tokTypes.newline) {
        this.tokens.splice(lastOffset, this.tokens.length);
      }
    }
  };

  /**
   * TODO: move this to documentation
   * in generators, the tokens that have to be manually defined are:
   * num
   *  * will need custom serialization logic too
   *  * value should store actual number value, base and original code
   * regexp
   *  * will need custom serialization logic too
   *  * value should store the literal, flags and original code
   * string
   *  * will need custom serialization logic too
   *  * value should store actual string value, quote types, and original code
   * name
   *  * value should store the parsed name and the original code
   * tab
   *  * todo: decide if length should be stored on tokenization, or generated
   *  during serialization
   * indent
   * dedent
   * whitespace (should only be emitted by catchup code)
   * newline
   * blockCommentBody
   * lineCommentBody
   */

  TacoBuffer._massageTokenState = function _massageTokenState(state) {
    if (_lodashLangIsString2["default"](state)) {
      state = _horchata.Token.stateFromCode(state);
    } else if (_lodashLangIsString2["default"](state.type)) {
      state.type = _horchata.tokTypes[state.type];
    }
    return state;
  };

  TacoBuffer.prototype._push = function _push(state) {
    state = TacoBuffer._massageTokenState(state);
    // avoid redundant mapping marks
    if (this._isRedundantMappingMark(state)) {
      return;
    }

    this._insertForceSpace(state) || this._insertFormattingSpace(state);

    if (state.type === _horchata.tokTypes.newline) {
      this._insertIndentTokens();
    } else if (this.isLastType(_horchata.tokTypes.newline)) {
      this.tokens.push(new _horchata.Token(_horchata.tokTypes.tab, this._indent));
    }
    // TODO: ensure leading tab tokens, unless in parenthetized expression context
    this.tokens.push(_horchata.Token.fromState(state));
  };

  TacoBuffer.prototype._insertIndentTokens = function _insertIndentTokens() {
    if (this._indent !== this._lastIndent) {
      if (this._indent > this._lastIndent) {
        for (var i = this._indent - this._lastIndent; i > 0; i--) {
          this.tokens.push(new _horchata.Token(_horchata.tokTypes.indent));
        }
      } else {
        for (var i = this._lastIndent - this._indent; i > 0; i--) {
          this.tokens.push(new _horchata.Token(_horchata.tokTypes.dedent));
        }
      }
      this._lastIndent = this._indent;
    }
  };

  // TODO: reduce duplication between force space and formatting space.

  TacoBuffer.prototype._insertForceSpace = function _insertForceSpace(state) {
    var force = state.type.forceSpaceWhenAfter;
    if (!force) {
      return false;
    }
    var last = this._last();
    if (last == null) {
      return false;
    }
    var isForced = force[last.type.key] || last.type.keyword && force.keyword;
    if (!isForced) {
      return false;
    }

    if (isForced === true || isForced(last, state)) {
      this.tokens.push(new _horchata.Token(_horchata.tokTypes.whitespace, { code: ' ' }));
      return true;
    }
    return false;
  };

  TacoBuffer.prototype._insertFormattingSpace = function _insertFormattingSpace(state) {
    if (this.format.compact || this.format.preserve) return false;
    var last = this._last();
    var insertFormatting = last && last.type.formattingSpaceAfter;
    if (insertFormatting === true && state.type === _horchata.tokTypes.newline) insertFormatting = false;
    if (!insertFormatting) {
      var formatting = state.type.formattingSpaceWhenAfter;
      if (!formatting) {
        return false;
      }
      if (last == null) {
        return false;
      }
      insertFormatting = formatting[last.type.key] || last.type.keyword && formatting.keyword;
    }
    if (!insertFormatting) {
      return false;
    }

    if (insertFormatting === true || insertFormatting(last, state)) {
      this.tokens.push(new _horchata.Token(_horchata.tokTypes.whitespace, { code: ' ' }));
      return true;
    }
    return false;
  };

  TacoBuffer.prototype._isRedundantMappingMark = function _isRedundantMappingMark(state) {
    if (this.tokens.length <= 0) return false;
    var last = this.tokens[this.tokens.length - 1];

    if (state.type === _horchata.tokTypes.mappingMark && last.type === _horchata.tokTypes.mappingMark) {
      if (state.value.pos === last.value.pos) {
        return true;
      }
    }
    return false;
  };

  /**
   * Serialization
   */

  TacoBuffer.prototype.stringify = function stringify() {
    var _arr2 = this.tokens;

    for (var _i2 = 0; _i2 < _arr2.length; _i2++) {
      var token = _arr2[_i2];
      this._serialize(token);
    }
    return this.output;
  };

  TacoBuffer.prototype.preview = function preview() {
    return this.tokens.join('');
  };

  TacoBuffer.prototype._serialize = function _serialize(token) {
    var code = undefined,
        origLoc = undefined;
    if (token.type === _horchata.tokTypes.mappingMark) {
      this.mark(token.value.loc);
      return;
    }
    code = token.type.toCode(token, this);
    origLoc = token.origLoc || token.loc;
    if (origLoc) this.mark(origLoc.start);
    this.output += code;
    this.position.push(code);
    if (origLoc) this.mark(origLoc.end);
  };

  // Warnings

  TacoBuffer.prototype.onWarning = function onWarning(warning) {
    // default; can be overwritten by opts
    // console.warn(warning);
    this._warnings.push(warning);
  };

  return TacoBuffer;
})();

exports["default"] = TacoBuffer;
module.exports = exports["default"];